<?php
/***************************************************************
*  Copyright notice
*
*  (c) 2013 IMIA net based solutions (d.frerich@imia.de)
*  All rights reserved
*
*  This script is part of the TYPO3 project. The TYPO3 project is
*  free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  The GNU General Public License can be found at
*  http://www.gnu.org/copyleft/gpl.html.
*
*  This script is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  This copyright notice MUST APPEAR in all copies of the script!
***************************************************************/

namespace IMIA\BaseExt\Utility;

use Symfony\Component\Yaml\Yaml;
use Doctrine\Common\Annotations\AnnotationReader;
use IMIA\BaseExt\Metadata\Driver\AnnotationDriver;
use IMIA\BaseExt\Utility\Generate\SQL;
use IMIA\BaseExt\Utility\Generate\TCA;
use IMIA\BaseExt\Utility\Generate\TypoScript;

/**
 * @package     imia_base_ext
 * @subpackage  Utility
 * @author      David Frerich <d.frerich@imia.de>
 */
class Generate
{
    /**
     * @var Cache
     */
    static public $cache;

    /**
     * @var string
     */
    static public $cacheKey;

    /**
     * @var boolean
     */
    static private $create = false;

    /**
     * @var array
     */
    static private $settings = array();

    /**
     * @var array
     */
    static private $metadata = array();

    /**
     * @var array
     */
    static private $autoload = array();

    /**
     * @var array
     */
    static private $ignoreAnnotations = array('validate', 'lazy', 'inject', 'fixme', 'todo');

    /**
     * Build the sql and tca for each extension
     */
    static public function build()
    {
        self::load();

        if (self::$create) {
            foreach (self::$settings as $extKey => $settings)
            {
                $updatedSQL = self::createSql($extKey);
                $updatedTCA = self::createTca($extKey);
                $updatedMapping = self::createTypoScript($extKey);

                if ($updatedSQL || $updatedTCA || $updatedMapping) {
                    self::updateTypo3LoadedExtensionInformation();
                }
                self::createAutoloading($extKey);
            }

            if (count(self::$autoload)) {
                self::updateAutoloadingCache(self::$autoload);

                // reload after generation of new autoload
                header('Location: ' . \TYPO3\CMS\Core\Utility\GeneralUtility::locationHeaderUrl($_SERVER['REQUEST_URI']));
            }
        }
    }

    /**
     * Clear the cached sql and tca for each extension
     */
    static public function clearCache()
    {
        self::init();

        self::$cache->resetCache('cache_core/ext_settings_' . self::$cacheKey . '.php');
        self::$cache->resetCache('cache_sql/');
        self::$cache->resetCache('cache_tca/');
        self::$cache->resetCache('cache_ts/');exit;
    }

    /**
     * Load the extension settings
     */
    static private function init()
    {
        self::$cacheKey = sha1(TYPO3_version . PATH_site . 'imia_base_ext');

        if (!self::$cache) {
            self::$cache = new Cache(PATH_site . 'typo3temp/Cache/Code/');
        }
    }

    /**
     * Load the extension settings
     */
    static private function load()
    {
        self::init();

        if (self::$cache->hasCache('cache_core/ext_settings_' . self::$cacheKey . '.php')) {
            self::$settings = self::$cache->loadCache('cache_core/ext_settings_' . self::$cacheKey . '.php');
        } else {
            foreach ($GLOBALS['TYPO3_LOADED_EXT'] as $extKey => $extInfo) {
                $extPath = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey);
                $settingsPath = $extPath . 'ext_baseext.yml';

                // deprecated
                if (!file_exists($settingsPath)) {
                    $settingsPath = $extPath . 'Configuration/ImiaBaseExt/settings.yml';
                }

                if (file_exists($settingsPath)) {
                    self::$settings[$extKey] = Yaml::parse($settingsPath);
                    self::$settings[$extKey]['path'] = $extPath;
                }
            }
            self::$cache->writeCache('cache_core/ext_settings_' . self::$cacheKey . '.php', self::$settings);
            self::$create = true;
        }
    }

    /**
     * @param string $extKey
     * @return boolean
     */
    static private function createSql($extKey)
    {
        $update = false;
        if (self::$settings[$extKey]['annotations']['sql'] != false) {
            $cacheFile = new File('ext_tables_' . File::flatten($extKey) . '.sql', self::$cache->getCachePath() . 'cache_sql/');

            if (!$cacheFile->exists()) {
                $metadata = self::loadMetadata($extKey);

                $sqlGen = new SQL($extKey);
                $sql = $sqlGen->build($metadata);

                $cacheFile->write($sql);
                $update = true;
            }

            $GLOBALS['TYPO3_LOADED_EXT'][$extKey]['ext_tables.sql'] = (string)$cacheFile;
        }

        return $update;
    }

    /**
     * @param string $extKey
     * @return boolean
     */
    static private function createTca($extKey)
    {
        $update = false;
        if (self::$settings[$extKey]['annotations']['tca'] != false) {
            $cacheFile = new File('ext_tables_' . File::flatten($extKey) . '_' . self::$cacheKey . '.php', self::$cache->getCachePath() . 'cache_tca/');

            if (!$cacheFile->exists()) {
                $cacheFile->setContents('<?php' . LF .
                    'if (!defined (\'TYPO3_MODE\')) {' . LF .
                    '	die (\'Access denied.\');' . LF .
                    '}' . LF);

                if (file_exists(self::$settings[$extKey]['path'] . 'ext_tables.php')) {
                    $cacheFile->addContents('@require_once(' .
                        str_replace('\'' . PATH_site, 'PATH_site . \'', '\'' . self::$settings[$extKey]['path'] . 'ext_tables.php\'') .
                        ');' . LF);
                }

                $metadata = self::loadMetadata($extKey);

                $tcaGen = new TCA($extKey);
                $cacheFiles = $tcaGen->build($metadata, $cacheFile);

                foreach ($cacheFiles as $file) {
                    $file->write();
                }

                $update = true;
            }

            $GLOBALS['TYPO3_LOADED_EXT'][$extKey]['ext_tables.php'] = (string)$cacheFile;
        }

        return $update;
    }

    /**
     * @param string $extKey
     * @return boolean
     */
    static private function createTypoScript($extKey)
    {
        $update = false;
        if (self::$settings[$extKey]['annotations']['sql'] != false) {
            $tsCacheFile = new File('ext_typoscript_setup_' . File::flatten($extKey) . '.ts', self::$cache->getCachePath() . 'cache_ts/');
            $cacheFile = new File('ext_localconf_' . File::flatten($extKey) . '_' . self::$cacheKey . '.php', self::$cache->getCachePath() . 'cache_ts/');

            if (!$tsCacheFile->exists() || !$cacheFile->exists()) {
                $cacheFile->setContents('<?php' . LF .
                    'if (!defined (\'TYPO3_MODE\')) {' . LF .
                    '	die (\'Access denied.\');' . LF .
                    '}' . LF);

                if (file_exists(self::$settings[$extKey]['path'] . 'ext_localconf.php')) {
                    $cacheFile->addContents('@require_once(' .
                        str_replace('\'' . PATH_site, 'PATH_site . \'', '\'' . self::$settings[$extKey]['path'] . 'ext_localconf.php\'') .
                        ');' . LF);
                }

                $metadata = self::loadMetadata($extKey);

                $tsGen = new TypoScript($extKey);
                $typoScript = $tsGen->build($metadata);

                $tsCacheFile->setContents($typoScript);
                $tsCacheFile->write();

                $cacheFile->addContents('\TYPO3\CMS\Core\Utility\ExtensionManagementUtility::addTypoScript($_EXTKEY, \'setup\',
                    file_get_contents(\'' . self::$cache->getCachePath() . 'cache_ts/ext_typoscript_setup_' . File::flatten($extKey) . '.ts' . '\'), true);');
                $cacheFile->write();
                $update = true;
            }

            $GLOBALS['TYPO3_LOADED_EXT'][$extKey]['ext_localconf.php'] = (string)$cacheFile;
        }

        return $update;
    }

    /**
     * @param string $extKey
     */
    static private function createAutoloading($extKey)
    {
        $extPath = \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::extPath($extKey);

        if (self::$settings[$extKey]['autoload'] != false || $extKey == 'imia_base_ext') {
            self::$autoload = array_merge(self::$autoload, self::getClassesFromPath($extPath . 'Classes'));
        }
    }

    static private function getClassesFromPath($path)
    {
        $classes = array();
        if (is_dir($path)) {
            $dir = dir($path);
            while (false !== ($entry = $dir->read())) {
                if (substr($entry, strlen($entry)-4) == '.php') {
                    $classes[strtolower(self::getClassFromPath($path . '/' . $entry))] = $path . '/' . $entry;
                } elseif ($entry != '.' && $entry != '..') {
                    $classes = array_merge($classes, self::getClassesFromPath($path . '/' . $entry));
                }
            }
            $dir->close();
        }

        return $classes;
    }

    /**
     * Update the typo3 loaded extension information array cache
     */
    static private function updateTypo3LoadedExtensionInformation()
    {
        \TYPO3\CMS\Core\Utility\ExtensionManagementUtility::removeCacheFiles();

        $cacheIdentifier = 'loaded_extensions_' . sha1((TYPO3_version . PATH_site . 'loadedExtensions'));
        $codeCache = $GLOBALS['typo3CacheManager']->getCache('cache_core');
        $codeCache->set($cacheIdentifier, 'return ' . var_export($GLOBALS['TYPO3_LOADED_EXT'], TRUE) . ';');
    }


	/**
	 * Set or update class loader cache entry.
     *
     * @param array $additionalClassRegistry
	 */
	static private function updateAutoloadingCache($additionalClassRegistry)
    {
        \TYPO3\CMS\Core\Core\ClassLoader::unregisterAutoloader();

        $identifier = 'ClassLoader_' . sha1((TYPO3_version . PATH_site . 'ClassLoader'));

        $phpCodeCache = $GLOBALS['typo3CacheManager']->getCache('cache_core');
		if ($phpCodeCache->has($identifier)) {
			list($classRegistry, $aliasToClassNameMapping) = require(PATH_site . 'typo3temp/Cache/Code/cache_core/' . $identifier . '.php');
		}
        $classRegistry = array_merge($classRegistry, $additionalClassRegistry);

		$cachedFileContent = 'return ' . var_export(array($classRegistry, $aliasToClassNameMapping), true) . ';';
		$GLOBALS['typo3CacheManager']->getCache('cache_core')->set($identifier, $cachedFileContent);
	}

    /**
     * @param string $extKey
     * @return array
     */
    static private function loadMetadata($extKey)
    {
        if (!array_key_exists($extKey, self::$metadata)) {
            $classes = self::loadClasses(self::$settings[$extKey]['path'] . '/Classes/Domain/Model');
            $classNames = array();
            foreach ($classes as $classPath) {
                require_once($classPath);
                $classNames[] = self::getClassFromPath($classPath);
            }

            $reader = new AnnotationReader();
            foreach (self::$ignoreAnnotations as $annotation) {
                $reader::addGlobalIgnoredName(strtolower($annotation));
                $reader::addGlobalIgnoredName(strtoupper($annotation));
            }

            $annotationDriver = new AnnotationDriver($reader);
            $metadataFactory = new \Metadata\MetadataFactory($annotationDriver);

            self::$metadata[$extKey] = array();
            foreach ($classNames as $className) {
                self::$metadata[$extKey][$className] = $metadataFactory->getMetadataForClass($className);
            }
        }

        return self::$metadata[$extKey];
    }

    /**
     * @param string $path
     * @return array
     */
    static private function loadClasses($path)
    {
        $classes = array();

        if (is_dir($path)) {
            $dir = dir($path);
            while (false !== ($entry = $dir->read())) {
                if (!in_array($entry, array('.', '..'))) {
                    $entryPath = realpath($path . '/' . $entry);
                    if (is_dir($entryPath)) {
                        $classes = array_merge($classes, self::loadClasses($entryPath));
                    } elseif (preg_match('/\.php$/ism', $entry) && is_file($entryPath)) {
                        $classes[] = $entryPath;
                    }
                }
            }
            $dir->close();
        }

        return $classes;
    }

    /**
     * @param string $path
     * @return string
     */
    static private function getClassFromPath($path)
    {
        $className = self::getClassFromPathByToken($path);

        if (!$className) {
            die('No class found in: ' . $path);
        } elseif (substr($className, strlen($className)-1) == '\\') {
            $className .= preg_replace('/^.*?([^\/]+)\.php$/ism', '$1', $path);
        }

        return $className;
    }

    /**
     * @param string $path
     * @return string
     */
    static private function getClassFromPathByToken($path)
    {
        $class = '';
        $namespace = '';
        $buffer = '';

        $fp = fopen($path, 'r');
        $i = 0;
        while (!feof($fp) && !$class) {
            $buffer .= fread($fp, 512);
            $tokens = @token_get_all($buffer);

            if (strpos($buffer, '{') === false) continue;

            for (;$i<count($tokens);$i++) {
                if ($tokens[$i][0] === T_NAMESPACE) {
                    for ($j=$i+1;$j<count($tokens); $j++) {
                        if ($tokens[$j][0] === T_STRING) {
                             $namespace .= '\\'.$tokens[$j][1];
                        } else if ($tokens[$j] === '{' || $tokens[$j] === ';') {
                             break;
                        }
                    }
                }

                if ($tokens[$i][0] === T_CLASS) {
                    for ($j=$i+1;$j<count($tokens);$j++) {
                        if ($tokens[$j] === '{') {
                            $class = $tokens[$i+2][1];
                        }
                    }
                }
            }
        }

        return substr($namespace, 1) . '\\' . $class;
    }
}